/*
 * Copyright (c) 2022 Meltytech, LLC
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef QMLRICHTEXTMENU_H
#define QMLRICHTEXTMENU_H

#include <QObject>

class QmlRichTextMenu : public QObject
{
    Q_OBJECT

public:
    explicit QmlRichTextMenu(QObject *parent = 0);

signals:
    void openTriggered();
    void saveAsTriggered();
    void undoTriggered();
    void redoTriggered();
    void cutTriggered();
    void copyTriggered();
    void pasteTriggered();
    void pastePlainTriggered();
    void selectAllTriggered();
    void insertTableTriggered();

public slots:
    void popup();
};

#endif // QMLRICHTEXTMENU_H
